;;;; make-snytax-case-extensions.scm - Generate expansions of the macros defined in `chicken-highlevel-macros.scm' - felix
;
; Copyright (c) 2000-2004, Felix L. Winkelmann
; All rights reserved.
;
; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
; conditions are met:
;
;   Redistributions of source code must retain the above copyright notice, this list of conditions and the following
;     disclaimer. 
;   Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following
;     disclaimer in the documentation and/or other materials provided with the distribution. 
;   Neither the name of the author nor the names of its contributors may be used to endorse or promote
;     products derived from this software without specific prior written permission. 
;
; THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS
; OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
; AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR
; CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
; CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
; SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
; THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
; OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
; POSSIBILITY OF SUCH DAMAGE.
;
; Send bugs, suggestions and ideas to: 
;
; felix@call-with-current-continuation.org
;
; Felix L. Winkelmann
; Unter den Gleichen 1
; 37130 Gleichen
; Germany


(define-values (*unit-name* *expansion-file* *output-file*)
  (apply values (command-line-arguments)) )

(define *expansion-output*)

(define eval
  (let ([eval eval])
    (lambda (x . env)
      (apply 
       eval 
       (if (and-let* ([(pair? x)] [a (car x)] [(string? a)])
	     (string=? "noexpand" a) )
	   (begin
	     (pp (cadr x) *expansion-output*)
	     (cadr x)  )
	   x)
       env) ) ) )

(define ##syncase#current-export-prefix #f)
(define ##syncase#properties (make-vector 3001 '()))

(define (putprop s k x)
  (let ([props (##sys#hash-table-ref ##syncase#properties s)])
    (if props
	(let ([a (assq props k)])
	  (if a
	      (##sys#setslot a 1 x)
	      (##sys#hash-table-set! ##syncase#properties s (cons (cons k x) props)) ) )
	(##sys#hash-table-set! ##syncase#properties s (list (cons k x))) ) ) )

(define (getprop s k)
  (and-let* ([props (##sys#hash-table-ref ##syncase#properties s)]
	     [a (assq k props)] )
    (cdr a) ) )

(define (remprop s k)
  (and-let* ([props (##sys#hash-table-ref ##syncase#properties s)])
    (let loop ([ps props] [prev #f])
      (unless (null? ps)
	(let ([a (##sys#slot ps 0)]
	      [r (##sys#slot ps 1)] )
	  (if (eq? k (##sys#slot a 0))
	      (if prev
		  (##sys#setslot prev 1 r)
		  (##sys#hash-table-set! ##syncase#properties s r) )
	      (loop r ps) ) ) ) ) ) )

(print "bootstrapping psyntax.pp ...")
(load "psyntax.pp")

(print "writing " *output-file* " ...")
(set! *expansion-output* (open-output-file *output-file*))

(fprintf
 *expansion-output*
#<<EOF
(declare
  (unit ~a)
  (disable-interrupts)
  (number-type fixnum)
  (usual-integrations) 
  (compress-literals 50 ##syncase#init-~a-literals) )

(define (##syncase#init-~a)

EOF
  *unit-name* *unit-name* *unit-name*)

(with-input-from-file *expansion-file*
  (lambda ()
    (let loop ()
      (let ([x (read)])
	(unless (eof-object? x)
	  (sc-expand x)
	  (loop))))))

(display ")\n" *expansion-output*)
(close-output-port *expansion-output*)
